/**
  *  \file h.cc
  *  \brief VGAP Hostside Lisp Interpreter
  *
  *  \author Stefan Reuther <Streu@gmx.de>
  *
  *  Share and enjoy.
  */

#include <iostream>
#include "phostpdk.h"
#include "lispy.h"
#include "lfile.h"
#include "lregexp.h"

/** Define integer symbol. */
static inline void
defineInt(const char* name, int value)
{
    LSymbol::define(name, new LInt(value));
}

/** Define enum.
    \param prefix      prefix for all values
    \param names       comma-separated list of enum names
    \param start_value value for first name */
static void
defineEnum(const char* prefix, const char* names, int start_value = 0)
{
    const char* p;
    while ((p = strchr(names, ',')) != 0) {
        std::string name(names, p);
        LSymbol::define(prefix + name, new LInt(start_value));
        names = p+1;
        start_value++;
    }
    LSymbol::define(std::string(prefix) + names, new LInt(start_value));
}

/** Convert null-terminated Uns16 array into list. */
static LPtr
listFromIdArray(const Uns16* ptr)
{
    LPtr n;
    LCons root(n, n);
    LCons* last = &root;
    while (*ptr) {
        LPtr val = new LInt(*ptr);
        LCons* nc = new LCons(val, LPtr());
        last->setCdr(nc);
        last = nc;
        ++ptr;
    }
    return root.getCdr();
}

static int
argSid(LPtr& p)
{
    int i = argInt(p);
    if (i <= 0 || i > SHIP_NR || !IsShipExist(i))
        throw LLispException("invalid-ship", "Ship Id out of range");
    return i;
}

static int
argPid(LPtr& p)
{
    int i = argInt(p);
    if (i <= 0 || i > PLANET_NR || !IsPlanetExist(i))
        throw LLispException("invalid-planet", "Planet Id out of range");
    return i;
}

static int
argBid(LPtr& p)
{
    int i = argInt(p);
    if (i <= 0 || i > PLANET_NR || !IsBaseExist(i))
        throw LLispException("invalid-base", "Base Id out of range");
    return i;
}

static int
argWid(LPtr& p)
{
    int i = argInt(p);
    if (i < 0 || i >= NumWormholes())
        throw LLispException("invalid-worm", "Wormhole Id out of range");
    return i;
}

static int
argMfid(LPtr& p)
{
    int i = argInt(p);
    if (i <= 0 || i > MINE_NR || !IsMinefieldExist(i))
        throw LLispException("invalid-minefield", "Minefield Id out of range");
    return i;
}

static RaceType_Def
argPlayer(LPtr& p)
{
    int i = argInt(p);
    if (i <= 0 || i > 11)
        throw LLispException("invalid-player", "player number out of range");
    return RaceType_Def(i);
}

static inline LPtr
intOrNull(int i)
{
    if (i)
        return new LInt(i);
    else
        return LPtr();
}

static inline LPtr
retBool(bool b)
{
    return b ? LPtr(&sym_t) : LPtr();
}

static void
prependSym(LPtr& p, const char* s)
{
    p = new LCons(new LSymbol(s), p);
}

//- (game-files-version)
//- => Returns a cons of (major . minor) version.
static LPtr
FGameFilesVersion(LPtr)
{
    Uns16 maj, min;
    GameFilesVersion(&maj, &min);
    return new LCons(new LInt(maj), new LInt(min));
}

//- (ship-functions SID)
//- => Returns list of special functions ship SID has.
//-    The list consists of symbols named like the hullfunc.txt names.
//-    In addition, two special symbols :heats and :cools are supported.
//-    E.g. (if (memq 'Cloak (ship-functions sid)) ...)
static LPtr
FShipFunctions(LPtr p)
{
    int sid = argSid(p);
    LPtr rv;
    if (ShipDoesAlchemy(sid))             prependSym(rv, "Alchemy");
    if (ShipDoesRefinery(sid))            prependSym(rv, "Refinery");
    if (ShipDoesAdvancedRefinery(sid))    prependSym(rv, "AdvancedRefinery");
    if (ShipCoolsTo50(sid))               prependSym(rv, "CoolsTo50");
    if (ShipHeatsTo50(sid))               prependSym(rv, "HeatsTo50");
    if (ShipHeatsTo100(sid))              prependSym(rv, "HeatsTo100");
    if (ShipCanHyperwarp(sid))            prependSym(rv, "Hyperdrive");
    if (ShipIsGravitonic(sid))            prependSym(rv, "Gravitonic");
    if (ShipScansAllWormholes(sid))       prependSym(rv, "ScansAllWormholes");
    if (ShipIsGamblingShip(sid))          prependSym(rv, "Gambling");
    if (ShipIsAntiCloaking(sid))          prependSym(rv, "AntiCloak");
    if (ShipDoesImperialAssault(sid))     prependSym(rv, "ImperialAssault");
    if (ShipDoesChunneling(sid))          prependSym(rv, "Chunneling");
    if (ShipHasRamScoop(sid))             prependSym(rv, "Ramscoop");
    if (ShipDoesFullBioscan(sid))         prependSym(rv, "FullBioscan");
    if (ShipHasAdvancedCloak(sid))        prependSym(rv, "AdvancedCloak");
    if (ShipCanCloak(sid))                prependSym(rv, "Cloak");
    if (ShipDoesBioscan(sid))             prependSym(rv, "Bioscan");
    if (ShipHasGloryDevice(sid))          prependSym(rv, "GloryDeviceLowDamage");
    if (ShipHasHEGloryDevice(sid))        prependSym(rv, "GloryDeviceHighDamage");
    if (ShipIsUnclonable(sid))            prependSym(rv, "Unclonable");
    if (ShipIsCloneableOnce(sid))         prependSym(rv, "CloneOnce");
    if (ShipIsUngiveable(sid))            prependSym(rv, "Ungiveable");
    if (ShipIsGiveableOnce(sid))          prependSym(rv, "GiveOnce");
    if (ShipHasLevel2Tow(sid))            prependSym(rv, "Level2Tow");
    if (ShipCanTow(sid))                  prependSym(rv, "Tow");
    if (ShipCanChunnelItself(sid))        prependSym(rv, "ChunnelSelf");
    if (ShipCanChunnelOthers(sid))        prependSym(rv, "ChunnelOthers");
    if (ShipCanBeChunnelMate(sid))        prependSym(rv, "ChunnelTarget");
    if (ShipIsImmuneToPlanetAttacks(sid)) prependSym(rv, "PlanetImmunity");
    if (ShipIsOreCondenser(sid))          prependSym(rv, "OreCondenser");
    if (ShipCanBoard(sid))                prependSym(rv, "Boarding");
    if (ShipIsImmuneToAntiCloak(sid))     prependSym(rv, "AntiCloakImmunity");
    if (ShipIsCrewAcademy(sid))           prependSym(rv, "Academy");
    if (ShipCanRepairOthers(sid))         prependSym(rv, "Repairs");
    if (ShipHasFullWeaponry(sid))         prependSym(rv, "FullWeaponry");
    if (ShipHasHardenedEngines(sid))      prependSym(rv, "HardenedEngines");

    if (ShipCools(sid))
        prependSym(rv, ":cools");
    if (ShipHeats(sid))
        prependSym(rv, ":heats");
    return rv;
}

//- (hull-functions HULL PLAYER)
//- => Returns list of special functions HULL has when built by PLAYER.
//-    See ship-functions for more.
static LPtr
FHullFunctions(LPtr p)
{
    int sid = argInt(p);
    int pid = argPlayer(p);
    if (sid <= 0 || sid > HULL_NR)
        throw LLispException("range-error", "Hull number out of range in 'hull-functions'");
    LPtr rv;
    if (HullDoesAlchemy(sid, pid))             prependSym(rv, "Alchemy");
    if (HullDoesRefinery(sid, pid))            prependSym(rv, "Refinery");
    if (HullDoesAdvancedRefinery(sid, pid))    prependSym(rv, "AdvancedRefinery");
    if (HullCoolsTo50(sid, pid))               prependSym(rv, "CoolsTo50");
    if (HullHeatsTo50(sid, pid))               prependSym(rv, "HeatsTo50");
    if (HullHeatsTo100(sid, pid))              prependSym(rv, "HeatsTo100");
    if (HullCanHyperwarp(sid, pid))            prependSym(rv, "Hyperdrive");
    if (HullIsGravitonic(sid, pid))            prependSym(rv, "Gravitonic");
    if (HullScansAllWormholes(sid, pid))       prependSym(rv, "ScansAllWormholes");
    if (HullIsGamblingShip(sid, pid))          prependSym(rv, "Gambling");
    if (HullIsAntiCloaking(sid, pid))          prependSym(rv, "AntiCloak");
    if (HullDoesImperialAssault(sid, pid))     prependSym(rv, "ImperialAssault");
    if (HullDoesChunneling(sid, pid))          prependSym(rv, "Chunneling");
    if (HullHasRamScoop(sid, pid))             prependSym(rv, "Ramscoop");
    if (HullDoesFullBioscan(sid, pid))         prependSym(rv, "FullBioscan");
    if (HullHasAdvancedCloak(sid, pid))        prependSym(rv, "AdvancedCloak");
    if (HullCanCloak(sid, pid))                prependSym(rv, "Cloak");
    if (HullDoesBioscan(sid, pid))             prependSym(rv, "Bioscan");
    if (HullHasGloryDevice(sid, pid))          prependSym(rv, "GloryDeviceLowDamage");
    if (HullHasHEGloryDevice(sid, pid))        prependSym(rv, "GloryDeviceHighDamage");
    if (HullIsUnclonable(sid, pid))            prependSym(rv, "Unclonable");
    if (HullIsCloneableOnce(sid, pid))         prependSym(rv, "CloneOnce");
    if (HullIsUngiveable(sid, pid))            prependSym(rv, "Ungiveable");
    if (HullIsGiveableOnce(sid, pid))          prependSym(rv, "GiveOnce");
    if (HullHasLevel2Tow(sid, pid))            prependSym(rv, "Level2Tow");
    if (HullCanTow(sid, pid))                  prependSym(rv, "Tow");
    if (HullCanChunnelItself(sid, pid))        prependSym(rv, "ChunnelSelf");
    if (HullCanChunnelOthers(sid, pid))        prependSym(rv, "ChunnelOthers");
    if (HullCanBeChunnelMate(sid, pid))        prependSym(rv, "ChunnelTarget");
    if (HullIsImmuneToPlanetAttacks(sid, pid)) prependSym(rv, "PlanetImmunity");
    if (HullIsOreCondenser(sid, pid))          prependSym(rv, "OreCondenser");
    if (HullCanBoard(sid, pid))                prependSym(rv, "Boarding");
    if (HullIsImmuneToAntiCloak(sid, pid))     prependSym(rv, "AntiCloakImmunity");
    if (HullIsCrewAcademy(sid, pid))           prependSym(rv, "Academy");
    if (HullCanRepairOthers(sid, pid))         prependSym(rv, "Repairs");
    if (HullHasFullWeaponry(sid, pid))         prependSym(rv, "FullWeaponry");
    if (HullHasHardenedEngines(sid, pid))      prependSym(rv, "HardenedEngines");

    if (HullCools(sid, pid))
        prependSym(rv, ":cools");
    if (HullHeats(sid, pid))
        prependSym(rv, ":heats");
    return rv;
}

static LPtr
doEnumObjects(int min, int max, Boolean IsExist(Uns16 id), RaceType_Def Owner(Uns16 id), LPtr p)
{
    int arg = p.nonnull() ? argInt(p) : 0;

    LPtr n;
    LCons root(n, n);
    LCons* last = &root;
    for (int i = min; i <= max; ++i) {
        if (IsExist(i) && (arg == 0 || Owner(i) == arg)) {
            LPtr val = new LInt(i);
            LCons* nc = new LCons(val, LPtr());
            last->setCdr(nc);
            last = nc;
        }
    }
    return root.getCdr();
}

//- (all-ships [plid])
//- => Returns a list of all existing ships. If plid is specified,
//-    only return ships owned by that player.
static LPtr
FAllShips(LPtr p)
{
    return doEnumObjects(1, SHIP_NR, IsShipExist, ShipOwner, p);
}

//- (all-planets [plid])
//- => Returns a list of all existing planets. If plid is specified,
//-    only return planets owned by that player.
static LPtr
FAllPlanets(LPtr p)
{
    return doEnumObjects(1, PLANET_NR, IsPlanetExist, PlanetOwner, p);
}

//- (all-bases [plid])
//- => Returns a list of all existing starbases. If plid is specified,
//-    only return bases owned by that player.
static LPtr
FAllBases(LPtr p)
{
    return doEnumObjects(1, BASE_NR, IsBaseExist, BaseOwner, p);
}

//- (all-minefields [plid])
//- => Return a list of all existing minefields. If plid is specified,
//-    only return minefields owned by that player.
static LPtr
FAllMinefields(LPtr p)
{
    return doEnumObjects(1, MINE_NR, IsMinefieldExist, MinefieldOwner, p);
}

#include "h.inc"

/** Initialize the symbols defined here. */
void
initPlanetsSymbols()
{
    /* Forward constants and enums from phostpdk.h into Lisp */
    defineInt("RACE-NR", RACE_NR);
    defineInt("SHIP-NR", SHIP_NR);
    defineInt("PLANET-NR", PLANET_NR);
    defineInt("BASE-NR", BASE_NR);
    defineInt("MINE-NR", MINE_NR);
    defineInt("HULL-NR", HULL_NR);
    defineInt("ENGINE-NR", ENGINE_NR);
    defineInt("BEAM-NR", BEAM_NR);
    defineInt("TORP-NR", TORP_NR);

    defineInt("LONGNAME-SIZE", LONGNAME_SIZE);
    defineInt("SHORTNAME-SIZE", SHORTNAME_SIZE);
    defineInt("ADJNAME-SIZE", ADJECTIVE_SIZE);
    defineInt("PLANETNAME-SIZE", PLANETNAME_SIZE);

    defineInt("RACEHULLS", RACEHULLS);
    defineInt("MAX-SPEED", MAX_SPEED);
    defineInt("MAX-BEAMS", MAX_BEAMS);
    defineInt("MAX-TUBES", MAX_TUBES);
    defineInt("MAX-BAYS", MAX_BAYS);
    defineInt("MAX-PLANET-BAYS", MAX_PLANET_BAYS);
    defineInt("MAX-BASE-DEFENSE", MAX_BASE_DEFENSE);
    defineInt("MAX-BASE-FIGHTERS", MAX_BASE_FIGHTERS);
    defineInt("MAX-TECH", MAX_TECH);
    defineInt("MAX-DENSITY", MAX_DENSITY);
    defineInt("MAX-HAPPY", MAX_HAPPY);
    defineInt("MIN-HAPPY", MIN_HAPPY);
    defineInt("MAX-COORDINATE", MAX_COORDINATE);
    defineInt("MIN-COORDINATE", MIN_COORDINATE);
    defineInt("MAX-SHIP-CREDITS", MAX_SHIP_CREDITS);
    defineInt("MAX-WORMHOLES", MAX_WORMHOLES);
    defineInt("MAX-MINES", MAX_MINES);
    defineInt("MAX-FACTORIES", MAX_FACTORIES);
    defineInt("MAX-DEFENSE", MAX_DEFENSE);
    defineInt("MAX-POPULATION", MAX_POPULATION);
    defineInt("MAX-TAX", MAX_TAX);
    defineInt("MAX-TEMP", MAX_TEMP);

    defineEnum("CARGO:", "NEUTRONIUM,TRITANIUM,DURANIUM,MOLYBDENUM,COLONISTS,SUPPLIES,CREDITS");
    defineEnum("RACE:", "NONE,FED,GORN,ROMULAN,KLINGON,ORION,BORG,THOLIAN,EMPIRE,CYLON,REBELS,COLONIES");
    defineEnum("NATIVE:", "NONE,HUMANOID,BOVINOID,REPTILIAN,AVIAN,AMORPHOUS,INSECTOID,AMPHIBIAN,GHIPSOLDAL,SILICONOID");
    defineEnum("GOVERNMENT:", "NONE,ANARCHY,PRE-TRIBAL,EARLY-TRIBAL,TRIBAL,FEUDAL,MONARCHY,REPRESENTATIVE,PARTICIPATORY,UNITY");
    defineEnum("MISSION:", "NONE,EXPLORE,MINE-SWEEP,LAY-MINES,KILL,SENSOR-SWEEP,COLONIZE,TOW,INTERCEPT,SPECIAL,CLOAK,"
               "GATHER-NEUTRONIUM,GATHER-DURANIUM,GATHER-TRITANIUM,GATHER-MOLYBDENUM,GATHER-SUPPLIES");
    defineEnum("BASE:", "NONE,REFUEL,MAX-DEFENSE,LOAD-TORPS,UNLOAD-FREIGHTERS,REPAIR-BASE,FORCE-SURRENDER");
    defineEnum("FIX:", "NONE,FIX,RECYCLE");
    defineEnum("TECH:", "ENGINE,HULL,BEAM,TORP");
    defineEnum("TONS:", "DESTROYED,DESTROYED-NOW,LOST,LOST-NOW");
    defineEnum("ALLY:", "SHIPS,PLANETS,MINES,COMBAT,VISION");
    defineEnum("ALLY-STATE:", "OFF,ON,CONDITIONAL");

    defineInt("OPEN:GAME-DIR-ONLY", GAME_DIR_ONLY);
    defineInt("OPEN:ROOT-DIR-ONLY", ROOT_DIR_ONLY);
    defineInt("OPEN:GAME-OR-ROOT-DIR", GAME_OR_ROOT_DIR);
    defineInt("OPEN:REWRITE-MODE", REWRITE_MODE);
    defineInt("OPEN:APPEND-MODE", APPEND_MODE);
    defineInt("OPEN:TEXT-MODE", TEXT_MODE);
    defineInt("OPEN:NO-MISSING-ERROR", NO_MISSING_ERROR);

    defineEnum("COMBAT:", "VICTOR,CAPTURED,DESTROYED,NOAMMO");
    defineEnum("LANG:", "ENGLISH,GERMAN,FRENCH,SPANISH,ITALIAN,DUTCH,RUSSIAN,ESTONIAN,NEW-ENGLISH,POLISH");
    defineEnum("BQ:", "PAL,FIFO,PBP");
    defineEnum("TS:", "No,Yes,Allies,External");
    defineEnum("UTIL:", "Ext,Dat,Tmp");
    defineEnum("POINTS:", "USED,UNUSED,ALL,THOST");
    defineEnum("SHIPS:", "CAPITAL,FREIGHTER,ALL");
    defineEnum("KILL:", "SHIPS,PLANETS,BASES,MINEFIELDS,STRUCTURES,CASH");

    defineInt("GAME-NAME-SIZE", GAME_NAME_SIZE);

    /* Register functions defined here */
    LSymbol::define("game-files-version", new LBuiltin(FGameFilesVersion, 0, 0));
    LSymbol::define("ship-functions", new LBuiltin(FShipFunctions, 1, 1));
    LSymbol::define("hull-functions", new LBuiltin(FHullFunctions, 2, 2));

    LSymbol::define("all-ships", new LBuiltin(FAllShips, 0, 1));
    LSymbol::define("all-planets", new LBuiltin(FAllPlanets, 0, 1));
    LSymbol::define("all-bases", new LBuiltin(FAllBases, 0, 1));
    LSymbol::define("all-minefields", new LBuiltin(FAllMinefields, 0, 1));

    /* Register the functions from h.def */
    initGeneratedFunctions();

    /* For reference, what follows are the PDK functions not yet implemented */
    //   extern Boolean WriteMessageToRST(RaceType_Def pRace, int pArgc,
    //         const char *pArgv[]);

    // /*! Single-read routines */
    //   extern IO_Def Read_AuxData_File(void);
    //   extern IO_Def Read_Turntime_File(void);
    //   extern IO_Def Read_HConfig_File(void);
    //   extern IO_Def Read_THost_HConfig_File(void);
    //   extern IO_Def Read_Planets_File(Int16 * pControl);
    //   extern IO_Def Read_Ships_File(Int16 * pControl);
    //   extern IO_Def Read_Bases_File(Int16 * pControl1, Int16 * pControl2);
    //   extern IO_Def Read_Mines_File(void);
    //   extern IO_Def Read_Hullspec_File(void);
    //   extern IO_Def Read_Truehull_File(void);
    //   extern IO_Def Read_Engspec_File(void);
    //   extern IO_Def Read_Torpspec_File(void);
    //   extern IO_Def Read_Beamspec_File(void);
    //   extern IO_Def Read_Planetname_File(void);
    //   extern IO_Def Read_Racenames_File(void);
    //   extern IO_Def Read_Xyplan_File(void);
    //   extern IO_Def Read_HostGen_File(void);
    //   extern void SpecialReadHullfunc(const char *shiplistFile, const char *hullfuncFile);

    // /*! Reporting */
    //   extern void Error(const char *pStr, ...);
    //   extern void ErrorExit(const char *pStr, ...);
    //   extern void Warning(const char *pStr, ...);
    //   extern void Info(const char *pStr, ...);

    // /*! Configuration file processing */
    //   typedef Boolean (*configAssignment_Func) (const char *lhs, char *rhs,
    //         const char *lInputLine);
    //   typedef void (*configError_Func) (const char*, ...);

    //   extern IO_Def ConfigFileReader(FILE * pInFile, const char *pFileName,
    //         const char *pSection, Boolean pUseDefaultSection,
    //         configAssignment_Func pAssignFunc);
    //   extern IO_Def ConfigFileReaderEx(FILE * pInFile, const char *pFileName, const char *pSection,
    //         Boolean pUseDefaultSection, configAssignment_Func pAssignFunc,
    //         configError_Func pError, Boolean pContinue);
    //     extern IO_Def Read_SomeConfig_File(FILE *pInFile, const char *pFileName,
    //         const char *pSection, Boolean pUseDefaultSection,
    //         Boolean clear, Boolean setDefaults);



    // /*
    //  *  Global variables
    //  */

    //   extern const char *gRootDirectory;
    //   extern const char *gGameDirectory;
    //   extern FILE *gLogFile;
    //   extern Boolean gNewlyMastered;
    //   extern Boolean gNonPHOSTWarnings;
    //   extern Boolean gConfigWarnings;
    //   extern Pconfig_Struct *gPconfigInfo;
    //   extern char *poptarg;
    //   extern int poptind;
    //   extern Boolean gUsingTHost;


    //   extern Boolean PutUtilRecordSimple(RaceType_Def pRace, Uns16 pRecordId,
    //                                      Uns16 pSize, void* pData);
    //   extern Boolean PutUtilFileTransfer(RaceType_Def pRace, const char* pName,
    //                                      Boolean pText, Uns16 pSize, void* pData);

    // /*
    //  *  points
    //  */
    //   extern Uns32 PointsFor(CargoType_Def lCargo);
    //   extern void PutPointsFor(CargoType_Def lCargo, Uns16 lPoints);
    //   extern Uns32 PointsForFighter(void);
    //   extern Uns32 PointsForBaseDefensePost(void);
    //   extern Uns32 PointsForMineralMine(void);
    //   extern Uns32 PointsForFactory(void);
    //   extern Uns32 PointsForDefencePost(void);
    //   extern Uns32 PointsForBaseOnPlanet(void);

    //   extern Uns32 PointsForHull(Uns16 lHullNr);
    //   extern Uns32 PointsForEngine(Uns16 lEngNr);
    //   extern Uns32 PointsForBeam(Uns16 lBeamNr);
    //   extern Uns32 PointsForTube(Uns16 lTorpNr);
    //   extern Uns32 PointsForTorpedo(Uns16 lTorpNr);
    //   extern Uns32 PointsForShipParts(Uns16 sID);
    //   extern Uns32 PointsForShipCargo(Uns16 sID);
    //   extern Uns32 PointsForShipAmmo(Uns16 sID);
    //   extern Uns32 PointsForPlanetStructures(Uns16 pID);
    //   extern Uns32 PointsForPlanetMinerals(Uns16 pID);
    //   extern Uns32 PointsForBaseTechLevels(Uns16 bID);
    //   extern Uns32 PointsForBaseDefense(Uns16 bID);
    //   extern Uns32 PointsForBaseStorage(Uns16 bID);
    //   extern Uns32 PointsForBaseTorpedos(Uns16 bID);
    //   extern Uns32 PointsForShip(Uns16 sID, PointsType_Def CountType);
    //   extern Uns32 PointsForPlanet(Uns16 pID, PointsType_Def CountType);
    //   extern Uns32 PointsForBase(Uns16 bID, PointsType_Def CountType);
    //   extern Uns32 PointsForMinefield(Uns16 mID, PointsType_Def CountType);
    //   extern Boolean IsShipType(Uns16 sID,ShipsType_Def ShipsType);
    //   extern Uns32 RaceScoreForShips( RaceType_Def Race, PointsType_Def CountType, ShipsType_Def ShipsType);
    //   extern Uns32 RaceScoreForPlanets( RaceType_Def Race, PointsType_Def CountType);
    //   extern Uns32 RaceScoreForBases( RaceType_Def Race, PointsType_Def CountType);
    //   extern Uns32 RaceScoreForMinefields( RaceType_Def Race, PointsType_Def CountType, Boolean RaceModify);
    //   extern Uns32 RaceShipsNumber( RaceType_Def Race, ShipsType_Def ShipsType);
    //   extern Uns32 RacePlanetsNumber(RaceType_Def Race);
    //   extern Uns32 RaceBasesNumber( RaceType_Def Race);
    //   extern Uns32 RaceFightersNumber(RaceType_Def Race);
    //   extern Uns32 RaceTorpedosNumber(RaceType_Def Race, Uns16 TorpType);
    //   extern Uns32 RaceBeamsNumber(RaceType_Def Race, Uns16 BeamType);
    //   extern Uns32 RaceFactoriesNumber(RaceType_Def Race);
    //   extern Uns32 RaceMineralMinesNumber(RaceType_Def Race);
    //   extern Uns32 RaceDefencePostsNumber(RaceType_Def Race);
    //   extern Uns32 RaceBaseDefenceNumber(RaceType_Def Race);
    //   extern Uns32 RaceCargoNumber(RaceType_Def Race,CargoType_Def CargoType);
    //   extern Uns32 RaceMinefieldUnitsNumber(RaceType_Def Race);
    //   extern Uns32 MineralsForHull(Uns16 lHullNr,  CargoType_Def CargoType);
    //   extern Uns32 MineralsForEngine(Uns16 lEngNr, CargoType_Def CargoType);
    //   extern Uns32 MineralsForBeam(Uns16 lBeamNr, CargoType_Def CargoType);
    //   extern Uns32 MineralsForTube(Uns16 lTorpNr, CargoType_Def CargoType);
    //   extern Uns32 MineralsForFighter(CargoType_Def CargoType);
    //   extern Uns32 MineralsForTorpedo(Uns16 lTorpNr, CargoType_Def CargoType);
    //   extern Uns32 MineralsForShipParts(Uns16 sID, CargoType_Def CargoType);
    //   extern Uns32 MineralsForShipCargo(Uns16 sID, CargoType_Def CargoType);
    //   extern Uns32 MineralsForShipAmmo(Uns16 sID, CargoType_Def CargoType);
    //   extern Uns32 MineralsForShip(Uns16 sID, PointsType_Def CountType, CargoType_Def CargoType);
    //   extern Uns32 MineralsForMineralMine(CargoType_Def CargoType);
    //   extern Uns32 MineralsForFactory(CargoType_Def CargoType);
    //   extern Uns32 MineralsForDefencePost(CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseOnPlanet(CargoType_Def CargoType);
    //   extern Uns32 MineralsForPlanetStructures(Uns16 pID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForPlanetMinerals(Uns16 pID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForPlanet(Uns16 pID, PointsType_Def CountType,CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseTechLevels(Uns16 bID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseDefensePost(CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseDefense(Uns16 bID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseStorage(Uns16 bID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForBaseTorpedos(Uns16 bID,CargoType_Def CargoType);
    //   extern Uns32 MineralsForBase(Uns16 bID, PointsType_Def CountType,CargoType_Def CargoType);
    //   extern Uns32 MineralsForMinefield(Uns16 mID, PointsType_Def CountType,CargoType_Def CargoType);
    //   extern Uns32 RaceMineralsForShips( RaceType_Def Race, PointsType_Def CountType, ShipsType_Def ShipsType, CargoType_Def CargoType);
    //   extern Uns32 RaceMineralsForPlanets( RaceType_Def Race, PointsType_Def CountType, CargoType_Def CargoType);
    //   extern Uns32 RaceMineralsForBases( RaceType_Def Race, PointsType_Def CountType, CargoType_Def CargoType);
    //   extern Uns32 RaceMineralsForMinefields( RaceType_Def Race, PointsType_Def CountType, Boolean RaceModify, CargoType_Def CargoType);


    //   /* friendly codes */
    //   extern void InitFCodes(void);
    //   extern void ShutdownFCodes(void);
    //   extern void DumpExtraFCodes(FILE *pFile);

    //   /* command messages */
    //   typedef Boolean (*CommandReader_Func) (Uns16 pRace,
    //                                          const char* pCommand,
    //                                          const char* pArgs,
    //                                          const char* pWholeLine,
    //                                          void* pData);
    //   typedef void (*CommandComplain_Func) (Uns16 pRace,
    //                                         const char* pLine,
    //                                         const char* pReason,
    //                                         void* pData);
    //   extern void CommandFileReader(Uns16 pRace,
    //                                 CommandReader_Func pFunc,
    //                                 CommandComplain_Func pComplain,
    //                                 const char* pProgName,
    //                                 const char* pPrivateFile,
    //                                 void* pData);
    //   extern void CommandFileReaderFP(Uns16 pRace,
    //                                   Uns16 pFileIsFor,
    //                                   CommandReader_Func pFunc,
    //                                   CommandComplain_Func pComplain,
    //                                   const char* pProgName,
    //                                   FILE* pFile,
    //                                   void* pData);
    //   extern void ComplainWithSubspaceMessage(Uns16 pRace, const char* pLine, const char* pReason, void* pData);
    //   extern void ComplainWithWarningMessage(Uns16 pRace, const char* pLine, const char* pReason, void* pData);
    //   extern char* GetToken(const char** pString, const char* pDelim);
}

int
main(int argc, char** argv)
{
    if (argc > 1)
        gGameDirectory = argv[1];
    if (argc > 2)
        gRootDirectory = argv[2];

    InitPHOSTLib();
    if (! ReadGlobalData())
        ErrorExit("Unable to read global data");
    if (! ReadHostData())
        ErrorExit("Unable to read host data");

    initBasicSymbols();
    initPlanetsSymbols();
    initFile();
    initRegexp();
    std::cout << "VGAP Host-Side Lisp Interpreter, compiled " __DATE__ "\n";
    std::cout << "(c) 2005,2006 Stefan Reuther <Streu@gmx.de>\n\n";
    if (argc > 3) {
        for (int i = 3; i < argc; ++i) {
            LPtr ex = new LCons(new LSymbol("load"),
                                new LCons(new LString(argv[i]),
                                          LPtr()));
            try {
                ex->eval();
                std::cout << "Loaded '" << argv[i] << "'.\n";
            }
            catch (LLispException& e) {
                std::cerr << "error: " << e.getData()->toString() << " (while loading '" << argv[i] << "')\n";
                break;
            }
        }
    } else {
        std::cout << "Enter Lisp expressions. Exit with EOF.\n";
        while (1) {
            try {
                LPtr x = parse(std::cin);
                std::cout << "=> " << x->eval()->toString() << "\n";
            }
            catch (LLispException& e) {
                if (e.getTag() == "eof")
                    break;

                std::cerr << "error: " << e.getTag() << " - " << e.getData()->toString() << "\n";
            }
        }
    }
    return 0;
}
